/*****************************************************************************/
/* Kernel SOCKS Bouncer (Kernel Patch) for 2.6.x kernels [KSB26]             */
/* (c) 2004, 2005 Paolo Ardoino <ardoino.gnu@disi.unige.it>                  */
/*****************************************************************************/
/*									     */
/* This program is free software; you can redistribute it and/or modify	     */
/* it under the terms of the GNU General Public License as published by	     */
/* the Free Software Foundation; either version 2 of the License, or	     */
/* (at your option) any later version.					     */
/* This program is distributed in the hope that it will be useful,	     */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     */
/* GNU General Public License for more details.				     */
/*									     */
/* You should have received a copy of the GNU General Public License	     */
/* along with this program; if not, write to the Free Software		     */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*****************************************************************************/

#include <linux/in.h>
#include <linux/inet.h>
#include <linux/byteorder/generic.h>
#include <linux/slab.h>
#include <linux/preempt.h>
#include <linux/fcntl.h>
#include <asm/semaphore.h>

#include "ksb26.h"

#define DEBUG 1 

#define KSB26_CONN_MAXTRIES 2

#define KSB26_SSEND do { oldfs = get_fs(); set_fs(get_ds()); \
	scall = sys_send(fd, sbuf, sblen, 0); \
	set_fs(oldfs); } while(0)
#define KSB26_SRECV do { scall = 0; oldfs = get_fs(); \
	set_fs(get_ds()); scall = sys_recv(fd, sbuf, PAGE_SIZE, 0); \
	set_fs(oldfs); } while(0)
#define KSB26_SHSOK do { if(sbuf[0] != 0x05 || sbuf[1] != 0x00) return -1; } while(0)
#define KSB26_SOCKSHS "\x05\x01%c"
#define KSB26_SOCKSRQ "\x05\x01%c\x03%c%s%c%c"
#define KSB26_DOWN_LISTSEM do { \
	down_interruptible(&ksb26_cntsem); \
	ksb26_connect_cnt++; \
	if(ksb26_connect_cnt == 1) \
		down_interruptible(&ksb26_listsem); \
	up(&ksb26_cntsem); } while(0)
#define KSB26_UP_LISTSEM do { \
	down_interruptible(&ksb26_cntsem); \
	ksb26_connect_cnt--; \
	if(ksb26_connect_cnt == 0) \
		up(&ksb26_listsem); \
	up(&ksb26_cntsem); } while(0)

asmlinkage long sys_connect_orig(int fd, struct sockaddr __user *uservaddr, int addrlen);

char ksb26;
unsigned int ksb26_nsocks;
unsigned int ksb26_nbhddrs;
struct ksb26_host ksb26_slh;
struct ksb26_host ksb26_bhlh;
struct semaphore ksb26_listsem;
struct semaphore ksb26_cntsem;
int ksb26_connect_cnt; 
int ksb26_maxsocks;

static char *ksb26_ntoa(unsigned long saddr)
{
	static char buff[18];
	char *p;

	p = (char *) &saddr;
	sprintf(buff, "%d.%d.%d.%d", (*p & 255), (*(p + 1) & 255), (*(p + 2) & 255), (*(p + 3) & 255));
	return buff;
}                                                        

static int ksb26_2bounce(unsigned long naddr)
{
	struct ksb26_host *ksb26_bhlaux = ksb26_bhlh.next;

	for(; ksb26_bhlaux != &ksb26_bhlh; ksb26_bhlaux = ksb26_bhlaux->next) {
		if(ksb26_bhlaux->naddr == naddr) 
			return 1;
	}
	return 0;
}

static int ksb26_sockshs(int fd, char *dhost, int dport)
{
	int scall;
	int sblen;
	mm_segment_t oldfs;
	char *sbuf;
  
	sbuf = (char *)kmalloc(PAGE_SIZE, GFP_KERNEL);
	sblen = sprintf(sbuf, KSB26_SOCKSHS, 0);
	KSB26_SSEND;
	KSB26_SRECV;
	KSB26_SHSOK;
  	sblen = sprintf(sbuf, KSB26_SOCKSRQ, 0, strlen(dhost), dhost, (dport >> 8) & 0xFF, dport & 0xFF);
	KSB26_SSEND;
	KSB26_SRECV;
	KSB26_SHSOK;
	kfree(sbuf);
	return scall;
}

asmlinkage long sys_connect_ksb26(int fd, struct sockaddr __user *uservaddr, int addrlen)
{
	mm_segment_t oldfs;
	struct sockaddr_in *ksin = (struct sockaddr_in *)uservaddr;
	struct ksb26_host *ksb26_slaux;
	unsigned long int daddr;
	unsigned short int dport;
	int err = 0, scnt;
	int sock_flags_old;
	
	KSB26_DOWN_LISTSEM;	
	oldfs = get_fs();
	set_fs(get_ds());
	if (ksb26_2bounce(ksin->sin_addr.s_addr) == 1) {
#ifdef DEBUG
		printk("[%s] SYS_CONNECT_KSB26 : localhost -> %s:%d\n", MODNAME, ksb26_ntoa(ksin->sin_addr.s_addr), ntohs(ksin->sin_port));
#endif
		ksb26_slaux = ksb26_slh.next;
		if(ksb26_nsocks == 0 || ksb26_slaux == &ksb26_slh) {
#ifdef DEBUG
			printk("[%s] No socks5 in list.\n", MODNAME);
#endif
			set_fs(oldfs);
			KSB26_UP_LISTSEM;
			return -1;
		}
		daddr = ksin->sin_addr.s_addr;
		dport = ksin->sin_port;
		if((sock_flags_old = sys_fcntl(fd, F_GETFL, 0)) < 0) {
			printk("[%s] Cannot read socket flags.\n", MODNAME);
			set_fs(oldfs);
			KSB26_UP_LISTSEM;
			return -1;
		}
		if(sock_flags_old != 2)
			sys_fcntl(fd, F_SETFL, 2);
		set_fs(oldfs);
		for(; ksb26_slaux != &ksb26_slh; ksb26_slaux = ksb26_slaux->next) {
#ifdef DEBUG
			printk("[%s] Trying [1] %s:%d socks5.\n", MODNAME, ksb26_slaux->ip, ksb26_slaux->port);
#endif
			oldfs = get_fs();
			set_fs(get_ds());
			ksin->sin_addr.s_addr = ksb26_slaux->naddr;
			ksin->sin_port = ksb26_slaux->nport;
			set_fs(oldfs);
			if((err = sys_connect_orig(fd, uservaddr, addrlen)) == 0)
				break;
		}
		if(err != 0) {
			KSB26_UP_LISTSEM;
			return err;
		}
#ifdef DEBUG
		printk("[%s] Connected to %s:%d [1] socks5.\n", MODNAME, ksb26_slaux->ip, ksb26_slaux->port);
#endif
		for(scnt = 1, ksb26_slaux = ksb26_slaux->next; ksb26_slaux != &ksb26_slh && scnt < ksb26_maxsocks ; ksb26_slaux = ksb26_slaux->next) {
			if(ksb26_slaux->wrk == 1) {
#ifdef DEBUG
				printk("[%s] Trying [%d] %s:%d socks5.\n", MODNAME, scnt + 1, ksb26_slaux->ip, ksb26_slaux->port);
#endif
				if((err = ksb26_sockshs(fd, ksb26_slaux->ip, ksb26_slaux->port)) > 0) {
					scnt++;
					printk("[%s] Connected to %s:%d [%d] socks5.\n", MODNAME, ksb26_slaux->ip, ksb26_slaux->port, scnt);
				} else {
					ksb26_slaux->wrk = 0;
					KSB26_UP_LISTSEM;
#ifdef DEBUG
					printk("[%s] Please retry.\n", MODNAME);
#endif
					return err;
				}
			}
			
		}
		if(scnt < ksb26_maxsocks) {
			printk("[%s] Too few working socks5.\n", MODNAME);
			KSB26_UP_LISTSEM;
			return err;
		}
		if((err = ksb26_sockshs(fd, ksb26_ntoa(daddr), ntohs(dport))) <= 0) {
			KSB26_UP_LISTSEM;
			return -ECONNREFUSED;
		}
#ifdef DEBUG
		printk("[%s] Connected to destination host [%s:%d].\n", MODNAME, ksb26_ntoa(daddr), ntohs(dport));
#endif
		oldfs = get_fs();
		set_fs(get_ds());
		if(sock_flags_old != 2)
			sys_fcntl(fd, F_SETFL, sock_flags_old);
		set_fs(oldfs);
	} else  {
		set_fs(oldfs);
		err = sys_connect_orig(fd, uservaddr, addrlen);
	}
	KSB26_UP_LISTSEM;
	return err;
}

EXPORT_SYMBOL(ksb26);
EXPORT_SYMBOL(ksb26_nsocks);
EXPORT_SYMBOL(ksb26_connect_cnt);
EXPORT_SYMBOL(ksb26_slh);
EXPORT_SYMBOL(ksb26_bhlh);
EXPORT_SYMBOL(ksb26_listsem);
EXPORT_SYMBOL(ksb26_cntsem);
EXPORT_SYMBOL(ksb26_maxsocks);
